
# THIS FILE IS GENERATED FROM wofry SETUP.PY
short_version = '1.0.21'
version = '1.0.21'
full_version = '1.0.21'
git_revision = '0398bd433bb9a6530a9502c1983e241f67ba8e9a'
release = True

if not release:
    version = full_version
    short_version += ".dev"
